/*
 * this is a dump program for .SYM files
 * A SYM file has the following format:
 * 	Paragraph of last paragraph in file (last para contains 00 00 verlow verhi).
 * 	NULL word
 * 	Segment of entry point
 * 	Number of constant symbols
 * 	Offset to lookup table from start of current para.
 * 	Number of segments (excluding constants)
 *      Para offset of first segment start
 * 	Unknown word
 * 	Module name
 * 	DB 0
 * 	[ Value Name ] for each symbol
 * 	
 * 
 * Each segment has this format:
 * At offset S in the file
 * 	DW Paragraph of start of next segment data, null for last segment
 * 	DW Number of symbols
 * 	DW x	- Name lookup table starts at S+x
 * 	DW Ordinal of segment
 * 	DW 4 DUP(0) ?Reserved?
 * 	DW Paragraph of line number info for segment
 * 	DW 255?
 * 	Segment name
 * 	[ Value name ] for each symbol
 */

#include <stdio.h>
#include <process.h>
#include <string.h>
#include <malloc.h>
#include <exc.h>
#include <dgb.h>

typedef unsigned int WORD;
typedef unsigned char BYTE;
typedef char *PSTR;

typedef struct NSTR {
   BYTE len;
   char s[1];
} NSTR;

typedef struct SYMHDR {
   WORD ParaLen;
   WORD xxx1;
   WORD EntrySeg;
   WORD Nconsts;
   WORD OffLUT;
   WORD Nsegs;
   WORD FirstSeg;
   BYTE xxx2;
   NSTR ModuleName;
} SYMHDR;

typedef struct SYMSEG {
   WORD ParaNext;
   WORD Nsyms;
   WORD OffLUT;
   WORD Ordinal;
   WORD Rsrv[4];
   WORD LineNumbers;
   WORD xxx3;
   NSTR SegName;
} SYMSEG;

typedef struct SYMBOL {
   WORD value;
   NSTR Name;
} SYMBOL;

typedef SYMHDR *PSYMHDR;
typedef SYMSEG *PSYMSEG;
typedef SYMBOL *PSYMBOL;

exception HIT_EOF = "Unexpected end of file";
exception FILE_TOO_LARGE = "Symbol file too large for memory";
exception NAME_TOO_LONG = "Name exceeds 80 character limit";

PSTR segfile;
PSYMHDR symhdr;

void *emalloc (WORD);
void MyRead (void *, WORD, WORD, FILE *);
char *GetNstr(NSTR *, char *);
void ShowFileContents(void);
WORD PrintSymbol (PSYMBOL);
WORD DumpSegment(WORD);

/*
 * Copy an NSTR to a char * ...
 */
char *GetNstr(n, s)
NSTR *n;
char *s;
{
   static char MyString[81];
   if (n->len > 80) exc_raise(NAME_TOO_LONG);
   if (s == NULL) s = MyString;
   strncpy(s, n->s, n->len);
   s[n->len] = 0;
   return s;
}

/*
 * Read from a file. This version checks the result.
 */
void MyRead (buffer, size, count, stream)
void *buffer;
WORD size, count;
FILE *stream;
{
   WORD res = (WORD) fread (buffer, size, count, stream);
   if (ferror(stream)) exc_raise(OP_FAILED);
   if (res < count) exc_raise (HIT_EOF);
}

/*
 * Load a named symbol file into memory. If a symbol file is too large to fit in memory
 * then the program will fail.
 * N.B. This code assumes that sizeof(WORD) == 2. I could put sizeof(WORD) all the way through it,
 * but I have to pick WORD's directly out of the file after I have loaded it.
 */
void LoadSym (name)
char *name;
{
   WORD paras, FileSize;
   FILE *infile;
   char ver[4];
   if ((infile = fopen(name, "rb")) == NULL)
   {
      perror(name);
      exit(1);
   }
   MyRead (&paras, 2, 1, infile);
   if (paras > 0x0fff) exc_raise(FILE_TOO_LARGE);
   FileSize = paras * 16;
   symhdr = (PSYMHDR) segfile = emalloc (FileSize);
   * (WORD *) segfile = paras;
   MyRead (segfile + 2, 1, FileSize - 2, infile);
   MyRead (ver, 4, 1, infile);
   printf("Used MAPSYM v%d.%02d\n", ver[3], ver[2]);
}

void main (argc, argv)
int argc;
char **argv;
{
   BEGIN
   if (argc != 2)
   {
      fprintf(stderr, "Usage: SYMDUMP <sym filename>\n");
      exit(1);
   }
   
   LoadSym(argv[1]);
   ShowFileContents();
   EXCEPTION
      WHEN(OTHERS)
	 printf("%s: Exception: %s\n", argv[0], (char *) exc_code);
   END
}

WORD PrintSymbol (p)
register PSYMBOL p;
{
   printf("%04x %s\n", p->value, GetNstr(&p->Name, NULL));
   return p->Name.len + sizeof(SYMBOL) - 1;
}

PSTR IsPlural(n)
WORD n;
{
   return (n == 1 ? "" : "s");
}

void ShowFileContents()
{
   PSTR ps;
   WORD Seg;
   WORD i;
   printf("Module %s\n", GetNstr(&symhdr->ModuleName, NULL));
   printf("%d segment%s\n", symhdr->Nsegs, IsPlural(symhdr->Nsegs));
   printf("Entry segment = %#04x\n", symhdr->EntrySeg);
   printf("%d constant%s\n", symhdr->Nconsts, IsPlural (symhdr->Nconsts));
   printf("%d %d\n", symhdr->xxx1, symhdr->xxx2);
   ps = (PSTR) &symhdr->ModuleName + symhdr->ModuleName.len + sizeof(NSTR);
   
   for (i = 0; i < symhdr->Nconsts; i++)
   {
      ps += PrintSymbol ((PSYMBOL) ps);
   }

   Seg = symhdr->FirstSeg;
   for (i = 0; i < symhdr->Nsegs; i++)
   {
      Seg = DumpSegment (Seg);
   }
}

/* Dump information about the segment starting at offset 'Para' in the file */
/* Returns next para to dump from */
WORD DumpSegment(Para)
WORD Para;
{
   WORD i;
   PSTR p;
   PSYMSEG sym = (PSYMSEG) (segfile + 16 * Para);
   printf("\nSegment %s @%04x", GetNstr(&sym->SegName, NULL), sym->Ordinal);
/*   if (sym->LineNumbers != 0) printf("(linenumbers)");*/
   printf("\n%d symbol%s\n", sym->Nsyms, IsPlural(sym->Nsyms));
   printf("%d %d %d %d %d\n", sym->Rsrv[0], sym->Rsrv[1], sym->Rsrv[2], sym->Rsrv[3], sym->xxx3);

   /* For each symbol read value/name and display them */
   p = (PSTR) &sym->SegName + sym->SegName.len + sizeof(NSTR) - 1;
   for (i = 0; i < sym->Nsyms; i++)
   {
      p += PrintSymbol ((PSYMBOL) p);
   }
   /* Return next para */
   return sym->ParaNext;
}

void *emalloc (size)
WORD size;
{
   char *p = malloc(size);
   if (p == NULL) exc_raise(OUT_OF_MEMORY);
   return p;
}
